from typing import Any, Union

from express_relay.models.base import IntString, UUIDString
from express_relay.models.evm import (
    Address,
    BidEvm,
    BidResponseEvm,
    BidStatusEvm,
    Bytes32,
    HexString,
    OpportunityDeleteEvm,
    OpportunityEvm,
    SignedMessageString,
    TokenAmount,
)
from express_relay.models.svm import (
    BidResponseSvm,
    BidStatusSvm,
    BidSvm,
    OpportunityDeleteSvm,
    OpportunitySvm,
    SvmTransaction,
)
from pydantic import BaseModel, Discriminator, Field, RootModel, Tag
from typing_extensions import Annotated, Literal

Bid = Union[BidEvm, BidSvm]


class BidStatusUpdate(BaseModel):
    """
    Attributes:
        id: The ID of the bid.
        bid_status: The current status of the bid.
    """

    id: UUIDString
    bid_status: Union[BidStatusEvm, BidStatusSvm]


BidResponse = Union[BidResponseEvm, BidResponseSvm]
BidResponseRoot = RootModel[BidResponse]


class OpportunityBidParams(BaseModel):
    """
    Attributes:
        amount: The amount of the bid in wei.
        nonce: The nonce of the bid.
        deadline: The unix timestamp after which the bid becomes invalid.
    """

    amount: IntString
    nonce: IntString
    deadline: IntString


class OpportunityBid(BaseModel):
    """
    Attributes:
        opportunity_id: The ID of the opportunity.
        amount: The amount of the bid in wei.
        executor: The address of the executor.
        permission_key: The permission key to bid on.
        signature: The signature of the bid.
        deadline: The unix timestamp after which the bid becomes invalid.
        nonce: The nonce of the bid.
    """

    opportunity_id: UUIDString
    amount: IntString
    executor: Address
    permission_key: HexString
    signature: SignedMessageString
    deadline: IntString
    nonce: IntString

    model_config = {
        "arbitrary_types_allowed": True,
    }


class OpportunityParamsV1(BaseModel):
    """
    Attributes:
        target_calldata: The calldata for the contract call.
        chain_id: The chain ID to bid on.
        target_contract: The contract address to call.
        permission_key: The permission key to bid on.
        buy_tokens: The tokens to receive in the opportunity.
        sell_tokens: The tokens to spend in the opportunity.
        target_call_value: The value to send with the contract call.
        version: The version of the opportunity.
    """

    target_calldata: HexString
    chain_id: str
    target_contract: Address
    permission_key: HexString
    buy_tokens: list[TokenAmount]
    sell_tokens: list[TokenAmount]
    target_call_value: IntString
    version: Literal["v1"]


class OpportunityParams(BaseModel):
    """
    Attributes:
        params: The parameters of the opportunity.
    """

    params: Union[OpportunityParamsV1] = Field(..., discriminator="version")


Opportunity = Union[OpportunityEvm, OpportunitySvm]
OpportunityRoot = RootModel[Opportunity]
OpportunityDelete = Union[OpportunityDeleteEvm | OpportunityDeleteSvm]
OpportunityDeleteRoot = RootModel[OpportunityDelete]


class SubscribeMessageParams(BaseModel):
    """
    Attributes:
        method: A string literal "subscribe".
        chain_ids: The chain IDs to subscribe to.
    """

    method: Literal["subscribe"]
    chain_ids: list[str]


class UnsubscribeMessageParams(BaseModel):
    """
    Attributes:
        method: A string literal "unsubscribe".
        chain_ids: The chain IDs to subscribe to.
    """

    method: Literal["unsubscribe"]
    chain_ids: list[str]


class PostBidMessageParamsEvm(BaseModel):
    """
    Attributes:
        method: A string literal "post_bid".
        amount: The amount of the bid in wei.
        target_calldata: The calldata for the contract call.
        chain_id: The chain ID to bid on.
        target_contract: The contract address to call.
        permission_key: The permission key to bid on.
    """

    method: Literal["post_bid"]
    amount: IntString
    target_calldata: HexString
    chain_id: str
    target_contract: Address
    permission_key: HexString


class PostOnChainBidMessageParamsSvm(BaseModel):
    """
    Attributes:
        method: A string literal "post_bid".
        chain_id: The chain ID to bid on.
        transaction: The transaction including the bid.
        slot: The minimum slot required for the bid to be executed successfully
              None if the bid can be executed at any recent slot
    """

    method: Literal["post_bid"]
    chain_id: str
    transaction: SvmTransaction
    slot: int | None


class PostSwapBidMessageParamsSvm(BaseModel):
    """
    Attributes:
        method: A string literal "post_bid".
        chain_id: The chain ID to bid on.
        transaction: The transaction including the bid.
        opportunity_id: The ID of the swap opportunity.
    """

    method: Literal["post_bid"]
    type: Literal["swap"]
    chain_id: str
    transaction: SvmTransaction
    opportunity_id: UUIDString


def get_discriminator_value(v: Any) -> str:
    if isinstance(v, dict):
        if "transaction" in v:
            return "svm"
        return "evm"
    if getattr(v, "transaction", None):
        return "svm"
    return "evm"


PostBidMessageParams = Annotated[
    Union[
        Annotated[PostBidMessageParamsEvm, Tag("evm")],
        Annotated[
            Union[PostOnChainBidMessageParamsSvm, PostSwapBidMessageParamsSvm],
            Tag("svm"),
        ],
    ],
    Discriminator(get_discriminator_value),
]


class PostOpportunityBidMessageParams(BaseModel):
    """
    Attributes:
        method: A string literal "post_opportunity_bid".
        opportunity_id: The ID of the opportunity.
        amount: The amount of the bid in wei.
        executor: The address of the executor.
        permission_key: The permission key to bid on.
        signature: The signature of the bid.
        deadline: The unix timestamp after which the bid becomes invalid.
        nonce: The nonce of the bid.
    """

    method: Literal["post_opportunity_bid"]
    opportunity_id: UUIDString
    amount: IntString
    executor: Address
    permission_key: HexString
    signature: SignedMessageString
    deadline: IntString
    nonce: IntString

    model_config = {
        "arbitrary_types_allowed": True,
    }


class ClientMessage(BaseModel):
    """
    Attributes:
        params: The parameters of the message.
    """

    params: Union[
        SubscribeMessageParams,
        UnsubscribeMessageParams,
        PostBidMessageParams,
        PostOpportunityBidMessageParams,
    ] = Field(..., discriminator="method")


class OpportunityAdapterConfig(BaseModel):
    """
    Attributes:
        chain_id: The chain ID.
        opportunity_adapter_factory: The address of the opportunity adapter factory contract.
        opportunity_adapter_init_bytecode_hash: The hash of the init bytecode of the opportunity adapter.
        permit2: The address of the permit2 contract.
        weth: The address of the WETH contract.
    """

    chain_id: int
    opportunity_adapter_factory: Address
    opportunity_adapter_init_bytecode_hash: Bytes32
    permit2: Address
    weth: Address



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
from ..program_id import PROGRAM_ID


class CheckPermissionAccounts(typing.TypedDict):
    sysvar_instructions: Pubkey
    permission: Pubkey
    router: Pubkey
    config_router: Pubkey
    express_relay_metadata: Pubkey


def check_permission(
    accounts: CheckPermissionAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["sysvar_instructions"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["permission"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["router"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["config_router"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["express_relay_metadata"],
            is_signer=False,
            is_writable=False,
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x9a\xc7\xe8\xf2`H\xc5\xec"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class InputJSON(typing.TypedDict):
    kind: typing.Literal["Input"]


class OutputJSON(typing.TypedDict):
    kind: typing.Literal["Output"]


@dataclass
class Input:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Input"

    @classmethod
    def to_json(cls) -> InputJSON:
        return InputJSON(
            kind="Input",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Input": {},
        }


@dataclass
class Output:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Output"

    @classmethod
    def to_json(cls) -> OutputJSON:
        return OutputJSON(
            kind="Output",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Output": {},
        }


FeeTokenKind = typing.Union[Input, Output]
FeeTokenJSON = typing.Union[InputJSON, OutputJSON]


def from_decoded(obj: dict) -> FeeTokenKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Input" in obj:
        return Input()
    if "Output" in obj:
        return Output()
    raise ValueError("Invalid enum object")


def from_json(obj: FeeTokenJSON) -> FeeTokenKind:
    if obj["kind"] == "Input":
        return Input()
    if obj["kind"] == "Output":
        return Output()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen("Input" / borsh.CStruct(), "Output" / borsh.CStruct())



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class ExpressRelayMetadataJSON(typing.TypedDict):
    admin: str
    relayer_signer: str
    fee_receiver_relayer: str
    split_router_default: int
    split_relayer: int
    swap_platform_fee_bps: int


@dataclass
class ExpressRelayMetadata:
    discriminator: typing.ClassVar = b"\xccK\x85\x07\xaf\xf1\x82\x0b"
    layout: typing.ClassVar = borsh.CStruct(
        "admin" / BorshPubkey,
        "relayer_signer" / BorshPubkey,
        "fee_receiver_relayer" / BorshPubkey,
        "split_router_default" / borsh.U64,
        "split_relayer" / borsh.U64,
        "swap_platform_fee_bps" / borsh.U64,
    )
    admin: Pubkey
    relayer_signer: Pubkey
    fee_receiver_relayer: Pubkey
    split_router_default: int
    split_relayer: int
    swap_platform_fee_bps: int

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["ExpressRelayMetadata"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["ExpressRelayMetadata"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["ExpressRelayMetadata"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "ExpressRelayMetadata":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = ExpressRelayMetadata.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            admin=dec.admin,
            relayer_signer=dec.relayer_signer,
            fee_receiver_relayer=dec.fee_receiver_relayer,
            split_router_default=dec.split_router_default,
            split_relayer=dec.split_relayer,
            swap_platform_fee_bps=dec.swap_platform_fee_bps,
        )

    def to_json(self) -> ExpressRelayMetadataJSON:
        return {
            "admin": str(self.admin),
            "relayer_signer": str(self.relayer_signer),
            "fee_receiver_relayer": str(self.fee_receiver_relayer),
            "split_router_default": self.split_router_default,
            "split_relayer": self.split_relayer,
            "swap_platform_fee_bps": self.swap_platform_fee_bps,
        }

    @classmethod
    def from_json(cls, obj: ExpressRelayMetadataJSON) -> "ExpressRelayMetadata":
        return cls(
            admin=Pubkey.from_string(obj["admin"]),
            relayer_signer=Pubkey.from_string(obj["relayer_signer"]),
            fee_receiver_relayer=Pubkey.from_string(obj["fee_receiver_relayer"]),
            split_router_default=obj["split_router_default"],
            split_relayer=obj["split_relayer"],
            swap_platform_fee_bps=obj["swap_platform_fee_bps"],
        )



from typing import List, Sequence, Tuple, TypedDict

import spl.token.instructions as spl_token
from express_relay.svm.generated.limo.accounts import Order
from express_relay.svm.generated.limo.instructions import TakeOrderArgs, take_order
from express_relay.svm.generated.limo.program_id import PROGRAM_ID
from express_relay.svm.token_utils import (
    create_associated_token_account_idempotent,
    get_ata,
)
from solana.rpc.async_api import AsyncClient
from solana.rpc.types import MemcmpOpts
from solders import system_program
from solders.instruction import Instruction
from solders.pubkey import Pubkey
from solders.system_program import TransferParams
from solders.sysvar import INSTRUCTIONS
from spl.token._layouts import MINT_LAYOUT
from spl.token.constants import TOKEN_PROGRAM_ID, WRAPPED_SOL_MINT

ESCROW_VAULT_SEED = b"escrow_vault"
GLOBAL_AUTH_SEED = b"authority"
EXPRESS_RELAY_MEATADATA_SEED = b"metadata"
EXPRESS_RELAY_CONFIG_ROUTER_SEED = b"config_router"
EVENT_AUTH_SEED = b"__event_authority"
INTERMEDIARY_OUTPUT_TOKEN_ACCOUNT_SEED = b"intermediary"


class OrderStateAndAddress(TypedDict):
    state: Order
    address: Pubkey


class WSOLInstructions(TypedDict):
    create_ixs: List[Instruction]
    fill_ixs: List[Instruction]
    close_ixs: List[Instruction]
    ata: Pubkey


class LimoClient:
    def __init__(self, connection: AsyncClient):
        self._connection = connection

    async def get_all_orders_state_and_address_with_filters(
        self, filters: List[MemcmpOpts], global_config: Pubkey
    ) -> List[OrderStateAndAddress]:
        filters.append(MemcmpOpts(offset=8, bytes=str(global_config)))
        programs = await self._connection.get_program_accounts(
            PROGRAM_ID,
            commitment=None,
            encoding="base64",
            data_slice=None,
            filters=filters,
        )

        return [
            {"state": Order.decode(value.account.data), "address": value.pubkey}
            for value in programs.value
        ]

    async def get_mint_decimals(self, mint: Pubkey) -> int:
        mint_account = await self._connection.get_account_info(mint)
        if mint_account.value is None:
            raise ValueError("Mint account not found")
        bytes_data = mint_account.value.data
        if len(bytes_data) != MINT_LAYOUT.sizeof():
            raise ValueError("Invalid mint size")

        decoded_data = MINT_LAYOUT.parse(bytes_data)
        decimals = decoded_data.decimals
        return decimals

    def get_ata_and_create_ixn_if_required(
        self,
        owner: Pubkey,
        token_mint_address: Pubkey,
        token_program_id: Pubkey,
        payer: Pubkey,
    ) -> Tuple[Pubkey, Sequence[Instruction]]:
        ata = get_ata(owner, token_mint_address, token_program_id)
        ix = create_associated_token_account_idempotent(
            payer, owner, token_mint_address, token_program_id
        )
        return ata, [ix]

    def get_init_if_needed_wsol_create_and_close_ixs(
        self, owner: Pubkey, payer: Pubkey, amount_to_deposit_lamports: int
    ) -> WSOLInstructions:
        """
        Returns necessary instructions to create, fill and close a wrapped SOL account.
        Creation instruction is idempotent.
        Filling instruction doesn't take into account the current WSOL balance.
        Closing instruction always closes the WSOL account and unwraps all WSOL back to SOL.
        Args:
            owner: Who owns the WSOL token account
            payer: Who pays for the instructions
            amount_to_deposit_lamports: Amount of lamports to deposit into the WSOL account
        """
        ata = get_ata(owner, WRAPPED_SOL_MINT, TOKEN_PROGRAM_ID)

        create_ixs = [
            create_associated_token_account_idempotent(
                payer, owner, WRAPPED_SOL_MINT, TOKEN_PROGRAM_ID
            )
        ]

        fill_ixs = []
        if amount_to_deposit_lamports > 0 and payer == owner:
            fill_ixs = [
                system_program.transfer(
                    TransferParams(
                        from_pubkey=owner,
                        to_pubkey=ata,
                        lamports=amount_to_deposit_lamports,
                    )
                ),
                spl_token.sync_native(
                    spl_token.SyncNativeParams(TOKEN_PROGRAM_ID, ata)
                ),
            ]

        close_ixs = []
        if payer == owner:
            close_ixs = [
                spl_token.close_account(
                    spl_token.CloseAccountParams(
                        program_id=TOKEN_PROGRAM_ID,
                        account=ata,
                        dest=owner,
                        owner=owner,
                    )
                )
            ]
        return WSOLInstructions(
            create_ixs=create_ixs, fill_ixs=fill_ixs, close_ixs=close_ixs, ata=ata
        )

    def take_order_ix(
        self,
        taker: Pubkey,
        order: OrderStateAndAddress,
        input_amount: int,
        output_amount: int,
        express_relay_program_id: Pubkey,
    ) -> List[Instruction]:
        """
        Returns the instructions to fulfill an order as a taker.
        Args:
            taker: The taker's public key
            order: The order to fulfill
            input_amount: The amount of input tokens to take.
            output_amount: The amount of output tokens to provide.
            express_relay_program_id: Express relay program id

        Returns:
            A list of instructions to include in the transaction to fulfill the order. The submit_bid instruction for
            express relay program is not included and should be added separately.

        """
        ixs: List[Instruction] = []
        close_wsol_ixns: List[Instruction] = []
        taker_input_ata: Pubkey
        if order["state"].input_mint == WRAPPED_SOL_MINT:
            instructions = self.get_init_if_needed_wsol_create_and_close_ixs(
                owner=taker, payer=taker, amount_to_deposit_lamports=0
            )
            ixs.extend(instructions["create_ixs"])
            close_wsol_ixns.extend(instructions["close_ixs"])
            taker_input_ata = instructions["ata"]
        else:
            (
                taker_input_ata,
                create_taker_input_ata_ixs,
            ) = self.get_ata_and_create_ixn_if_required(
                owner=taker,
                token_mint_address=order["state"].input_mint,
                token_program_id=order["state"].input_mint_program_id,
                payer=taker,
            )
            ixs.extend(create_taker_input_ata_ixs)

        taker_output_ata: Pubkey
        maker_output_ata: Pubkey | None = None
        intermediary_output_token_account: Pubkey | None = None
        if order["state"].output_mint == WRAPPED_SOL_MINT:
            instructions = self.get_init_if_needed_wsol_create_and_close_ixs(
                owner=taker, payer=taker, amount_to_deposit_lamports=output_amount
            )
            ixs.extend(instructions["create_ixs"])
            ixs.extend(instructions["fill_ixs"])
            close_wsol_ixns.extend(instructions["close_ixs"])
            taker_output_ata = instructions["ata"]

            intermediary_output_token_account = self.get_intermediary_token_account_pda(
                PROGRAM_ID, order["address"]
            )
        else:
            (
                taker_output_ata,
                create_taker_output_ata_ixs,
            ) = self.get_ata_and_create_ixn_if_required(
                owner=taker,
                token_mint_address=order["state"].output_mint,
                token_program_id=order["state"].output_mint_program_id,
                payer=taker,
            )
            ixs.extend(create_taker_output_ata_ixs)

            (
                maker_output_ata,
                create_maker_output_ata_ixs,
            ) = self.get_ata_and_create_ixn_if_required(
                owner=order["state"].maker,
                token_mint_address=order["state"].output_mint,
                token_program_id=order["state"].output_mint_program_id,
                payer=taker,
            )
            ixs.extend(create_maker_output_ata_ixs)

        pda_authority = self.get_pda_authority(PROGRAM_ID, order["state"].global_config)
        ixs.append(
            take_order(
                TakeOrderArgs(
                    input_amount=input_amount,
                    min_output_amount=output_amount,
                    tip_amount_permissionless_taking=0,
                ),
                {
                    "taker": taker,
                    "maker": order["state"].maker,
                    "global_config": order["state"].global_config,
                    "pda_authority": pda_authority,
                    "order": order["address"],
                    "input_mint": order["state"].input_mint,
                    "output_mint": order["state"].output_mint,
                    "input_vault": self.get_token_vault_pda(
                        PROGRAM_ID,
                        order["state"].global_config,
                        order["state"].input_mint,
                    ),
                    "taker_input_ata": taker_input_ata,
                    "taker_output_ata": taker_output_ata,
                    "intermediary_output_token_account": intermediary_output_token_account,
                    "maker_output_ata": maker_output_ata,
                    "express_relay": express_relay_program_id,
                    "express_relay_metadata": self.get_express_relay_metadata_pda(
                        express_relay_program_id
                    ),
                    "sysvar_instructions": INSTRUCTIONS,
                    "permission": order["address"],
                    "config_router": self.get_express_relay_config_router_pda(
                        express_relay_program_id, pda_authority
                    ),
                    "input_token_program": order["state"].input_mint_program_id,
                    "output_token_program": order["state"].output_mint_program_id,
                    "event_authority": self.get_event_authority(PROGRAM_ID),
                    "program": PROGRAM_ID,
                },
            )
        )

        ixs.extend(close_wsol_ixns)
        return ixs

    @staticmethod
    def get_program_id() -> Pubkey:
        return PROGRAM_ID

    @staticmethod
    def get_token_vault_pda(
        program_id: Pubkey, global_config: Pubkey, input_mint: Pubkey
    ) -> Pubkey:
        return Pubkey.find_program_address(
            seeds=[ESCROW_VAULT_SEED, bytes(global_config), bytes(input_mint)],
            program_id=program_id,
        )[0]

    @staticmethod
    def get_express_relay_metadata_pda(program_id: Pubkey) -> Pubkey:
        return Pubkey.find_program_address(
            seeds=[EXPRESS_RELAY_MEATADATA_SEED], program_id=program_id
        )[0]

    @staticmethod
    def get_express_relay_config_router_pda(
        program_id: Pubkey, router: Pubkey
    ) -> Pubkey:
        return Pubkey.find_program_address(
            seeds=[EXPRESS_RELAY_CONFIG_ROUTER_SEED, bytes(router)],
            program_id=program_id,
        )[0]

    @staticmethod
    def get_pda_authority(program_id: Pubkey, global_config: Pubkey) -> Pubkey:
        return Pubkey.find_program_address(
            seeds=[GLOBAL_AUTH_SEED, bytes(global_config)], program_id=program_id
        )[0]

    @staticmethod
    def get_event_authority(program_id: Pubkey) -> Pubkey:
        return Pubkey.find_program_address(
            seeds=[EVENT_AUTH_SEED], program_id=program_id
        )[0]

    @staticmethod
    def get_intermediary_token_account_pda(
        program_id: Pubkey, order_address: Pubkey
    ) -> Pubkey:
        return Pubkey.find_program_address(
            seeds=[INTERMEDIARY_OUTPUT_TOKEN_ACCOUNT_SEED, bytes(order_address)],
            program_id=program_id,
        )[0]



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class SwapArgs(typing.TypedDict):
    data: types.swap_args.SwapArgs


layout = borsh.CStruct("data" / types.swap_args.SwapArgs.layout)


class SwapAccounts(typing.TypedDict):
    searcher: Pubkey
    trader: Pubkey
    searcher_input_ta: Pubkey
    searcher_output_ta: Pubkey
    trader_input_ata: Pubkey
    trader_output_ata: Pubkey
    router_fee_receiver_ta: Pubkey
    relayer_fee_receiver_ata: Pubkey
    express_relay_fee_receiver_ata: Pubkey
    mint_input: Pubkey
    mint_output: Pubkey
    mint_fee: Pubkey
    token_program_input: Pubkey
    token_program_output: Pubkey
    token_program_fee: Pubkey
    express_relay_metadata: Pubkey


def swap(
    args: SwapArgs,
    accounts: SwapAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["searcher"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["trader"], is_signer=True, is_writable=False),
        AccountMeta(
            pubkey=accounts["searcher_input_ta"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["searcher_output_ta"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["trader_input_ata"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["trader_output_ata"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["router_fee_receiver_ta"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["relayer_fee_receiver_ata"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(
            pubkey=accounts["express_relay_fee_receiver_ata"],
            is_signer=False,
            is_writable=True,
        ),
        AccountMeta(pubkey=accounts["mint_input"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["mint_output"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["mint_fee"], is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["token_program_input"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["token_program_output"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["token_program_fee"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["express_relay_metadata"],
            is_signer=False,
            is_writable=False,
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xf8\xc6\x9e\x91\xe1u\x87\xc8"
    encoded_args = layout.build(
        {
            "data": args["data"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class InitializeArgsJSON(typing.TypedDict):
    split_router_default: int
    split_relayer: int


@dataclass
class InitializeArgs:
    layout: typing.ClassVar = borsh.CStruct(
        "split_router_default" / borsh.U64, "split_relayer" / borsh.U64
    )
    split_router_default: int
    split_relayer: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "InitializeArgs":
        return cls(
            split_router_default=obj.split_router_default,
            split_relayer=obj.split_relayer,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "split_router_default": self.split_router_default,
            "split_relayer": self.split_relayer,
        }

    def to_json(self) -> InitializeArgsJSON:
        return {
            "split_router_default": self.split_router_default,
            "split_relayer": self.split_relayer,
        }

    @classmethod
    def from_json(cls, obj: InitializeArgsJSON) -> "InitializeArgs":
        return cls(
            split_router_default=obj["split_router_default"],
            split_relayer=obj["split_relayer"],
        )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class SetSwapPlatformFeeArgsJSON(typing.TypedDict):
    swap_platform_fee_bps: int


@dataclass
class SetSwapPlatformFeeArgs:
    layout: typing.ClassVar = borsh.CStruct("swap_platform_fee_bps" / borsh.U64)
    swap_platform_fee_bps: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "SetSwapPlatformFeeArgs":
        return cls(swap_platform_fee_bps=obj.swap_platform_fee_bps)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"swap_platform_fee_bps": self.swap_platform_fee_bps}

    def to_json(self) -> SetSwapPlatformFeeArgsJSON:
        return {"swap_platform_fee_bps": self.swap_platform_fee_bps}

    @classmethod
    def from_json(cls, obj: SetSwapPlatformFeeArgsJSON) -> "SetSwapPlatformFeeArgs":
        return cls(swap_platform_fee_bps=obj["swap_platform_fee_bps"])



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class UpdateGlobalConfigArgs(typing.TypedDict):
    mode: int
    value: list[int]


layout = borsh.CStruct("mode" / borsh.U16, "value" / borsh.U8[128])


class UpdateGlobalConfigAccounts(typing.TypedDict):
    admin_authority: Pubkey
    global_config: Pubkey


def update_global_config(
    args: UpdateGlobalConfigArgs,
    accounts: UpdateGlobalConfigAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["admin_authority"], is_signer=True, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["global_config"], is_signer=False, is_writable=True
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xa4T\x82\xbdo:\xfa\xc8"
    encoded_args = layout.build(
        {
            "mode": args["mode"],
            "value": args["value"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class ExactInJSON(typing.TypedDict):
    kind: typing.Literal["ExactIn"]


class ExactOutJSON(typing.TypedDict):
    kind: typing.Literal["ExactOut"]


@dataclass
class ExactIn:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "ExactIn"

    @classmethod
    def to_json(cls) -> ExactInJSON:
        return ExactInJSON(
            kind="ExactIn",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "ExactIn": {},
        }


@dataclass
class ExactOut:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "ExactOut"

    @classmethod
    def to_json(cls) -> ExactOutJSON:
        return ExactOutJSON(
            kind="ExactOut",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "ExactOut": {},
        }


OrderTypeKind = typing.Union[ExactIn, ExactOut]
OrderTypeJSON = typing.Union[ExactInJSON, ExactOutJSON]


def from_decoded(obj: dict) -> OrderTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "ExactIn" in obj:
        return ExactIn()
    if "ExactOut" in obj:
        return ExactOut()
    raise ValueError("Invalid enum object")


def from_json(obj: OrderTypeJSON) -> OrderTypeKind:
    if obj["kind"] == "ExactIn":
        return ExactIn()
    if obj["kind"] == "ExactOut":
        return ExactOut()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen("ExactIn" / borsh.CStruct(), "ExactOut" / borsh.CStruct())



from __future__ import annotations
from . import (
    fee_token,
)
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class SwapArgsJSON(typing.TypedDict):
    deadline: int
    amount_input: int
    amount_output: int
    referral_fee_bps: int
    fee_token: fee_token.FeeTokenJSON


@dataclass
class SwapArgs:
    layout: typing.ClassVar = borsh.CStruct(
        "deadline" / borsh.I64,
        "amount_input" / borsh.U64,
        "amount_output" / borsh.U64,
        "referral_fee_bps" / borsh.U16,
        "fee_token" / fee_token.layout,
    )
    deadline: int
    amount_input: int
    amount_output: int
    referral_fee_bps: int
    fee_token: fee_token.FeeTokenKind

    @classmethod
    def from_decoded(cls, obj: Container) -> "SwapArgs":
        return cls(
            deadline=obj.deadline,
            amount_input=obj.amount_input,
            amount_output=obj.amount_output,
            referral_fee_bps=obj.referral_fee_bps,
            fee_token=fee_token.from_decoded(obj.fee_token),
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "deadline": self.deadline,
            "amount_input": self.amount_input,
            "amount_output": self.amount_output,
            "referral_fee_bps": self.referral_fee_bps,
            "fee_token": self.fee_token.to_encodable(),
        }

    def to_json(self) -> SwapArgsJSON:
        return {
            "deadline": self.deadline,
            "amount_input": self.amount_input,
            "amount_output": self.amount_output,
            "referral_fee_bps": self.referral_fee_bps,
            "fee_token": self.fee_token.to_json(),
        }

    @classmethod
    def from_json(cls, obj: SwapArgsJSON) -> "SwapArgs":
        return cls(
            deadline=obj["deadline"],
            amount_input=obj["amount_input"],
            amount_output=obj["amount_output"],
            referral_fee_bps=obj["referral_fee_bps"],
            fee_token=fee_token.from_json(obj["fee_token"]),
        )



from solana.constants import SYSTEM_PROGRAM_ID
from solders.instruction import AccountMeta, Instruction
from solders.pubkey import Pubkey
from solders.sysvar import RENT
from spl.token.constants import ASSOCIATED_TOKEN_PROGRAM_ID


def get_ata(
    owner: Pubkey, token_mint_address: Pubkey, token_program_id: Pubkey
) -> Pubkey:
    ata, _ = Pubkey.find_program_address(
        seeds=[bytes(owner), bytes(token_program_id), bytes(token_mint_address)],
        program_id=ASSOCIATED_TOKEN_PROGRAM_ID,
    )
    return ata


def create_associated_token_account_idempotent(
    payer: Pubkey, owner: Pubkey, mint: Pubkey, token_program_id: Pubkey
) -> Instruction:
    """Creates a transaction instruction to create an associated token account.

    Returns:
        The instruction to create the associated token account.
    """
    associated_token_address = get_ata(owner, mint, token_program_id)
    return Instruction(
        accounts=[
            AccountMeta(pubkey=payer, is_signer=True, is_writable=True),
            AccountMeta(
                pubkey=associated_token_address, is_signer=False, is_writable=True
            ),
            AccountMeta(pubkey=owner, is_signer=False, is_writable=False),
            AccountMeta(pubkey=mint, is_signer=False, is_writable=False),
            AccountMeta(pubkey=SYSTEM_PROGRAM_ID, is_signer=False, is_writable=False),
            AccountMeta(pubkey=token_program_id, is_signer=False, is_writable=False),
            AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
        ],
        program_id=ASSOCIATED_TOKEN_PROGRAM_ID,
        data=bytes([1]),  # idempotent version of the instruction
    )



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class SubmitBidArgsJSON(typing.TypedDict):
    deadline: int
    bid_amount: int


@dataclass
class SubmitBidArgs:
    layout: typing.ClassVar = borsh.CStruct(
        "deadline" / borsh.I64, "bid_amount" / borsh.U64
    )
    deadline: int
    bid_amount: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "SubmitBidArgs":
        return cls(deadline=obj.deadline, bid_amount=obj.bid_amount)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"deadline": self.deadline, "bid_amount": self.bid_amount}

    def to_json(self) -> SubmitBidArgsJSON:
        return {"deadline": self.deadline, "bid_amount": self.bid_amount}

    @classmethod
    def from_json(cls, obj: SubmitBidArgsJSON) -> "SubmitBidArgs":
        return cls(deadline=obj["deadline"], bid_amount=obj["bid_amount"])



import typing
from anchorpy.error import ProgramError


class OrderCanNotBeCanceled(ProgramError):
    def __init__(self) -> None:
        super().__init__(6000, "Order can't be canceled")

    code = 6000
    name = "OrderCanNotBeCanceled"
    msg = "Order can't be canceled"


class OrderNotActive(ProgramError):
    def __init__(self) -> None:
        super().__init__(6001, "Order not active")

    code = 6001
    name = "OrderNotActive"
    msg = "Order not active"


class InvalidAdminAuthority(ProgramError):
    def __init__(self) -> None:
        super().__init__(6002, "Invalid admin authority")

    code = 6002
    name = "InvalidAdminAuthority"
    msg = "Invalid admin authority"


class InvalidPdaAuthority(ProgramError):
    def __init__(self) -> None:
        super().__init__(6003, "Invalid pda authority")

    code = 6003
    name = "InvalidPdaAuthority"
    msg = "Invalid pda authority"


class InvalidConfigOption(ProgramError):
    def __init__(self) -> None:
        super().__init__(6004, "Invalid config option")

    code = 6004
    name = "InvalidConfigOption"
    msg = "Invalid config option"


class InvalidOrderOwner(ProgramError):
    def __init__(self) -> None:
        super().__init__(6005, "Order owner account is not the order owner")

    code = 6005
    name = "InvalidOrderOwner"
    msg = "Order owner account is not the order owner"


class OutOfRangeIntegralConversion(ProgramError):
    def __init__(self) -> None:
        super().__init__(6006, "Out of range integral conversion attempted")

    code = 6006
    name = "OutOfRangeIntegralConversion"
    msg = "Out of range integral conversion attempted"


class InvalidFlag(ProgramError):
    def __init__(self) -> None:
        super().__init__(6007, "Invalid boolean flag, valid values are 0 and 1")

    code = 6007
    name = "InvalidFlag"
    msg = "Invalid boolean flag, valid values are 0 and 1"


class MathOverflow(ProgramError):
    def __init__(self) -> None:
        super().__init__(6008, "Mathematical operation with overflow")

    code = 6008
    name = "MathOverflow"
    msg = "Mathematical operation with overflow"


class OrderInputAmountInvalid(ProgramError):
    def __init__(self) -> None:
        super().__init__(6009, "Order input amount invalid")

    code = 6009
    name = "OrderInputAmountInvalid"
    msg = "Order input amount invalid"


class OrderOutputAmountInvalid(ProgramError):
    def __init__(self) -> None:
        super().__init__(6010, "Order output amount invalid")

    code = 6010
    name = "OrderOutputAmountInvalid"
    msg = "Order output amount invalid"


class InvalidHostFee(ProgramError):
    def __init__(self) -> None:
        super().__init__(6011, "Host fee bps must be between 0 and 10000")

    code = 6011
    name = "InvalidHostFee"
    msg = "Host fee bps must be between 0 and 10000"


class IntegerOverflow(ProgramError):
    def __init__(self) -> None:
        super().__init__(6012, "Conversion between integers failed")

    code = 6012
    name = "IntegerOverflow"
    msg = "Conversion between integers failed"


class InvalidTipBalance(ProgramError):
    def __init__(self) -> None:
        super().__init__(6013, "Tip balance less than accounted tip")

    code = 6013
    name = "InvalidTipBalance"
    msg = "Tip balance less than accounted tip"


class InvalidTipTransferAmount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6014, "Tip transfer amount is less than expected")

    code = 6014
    name = "InvalidTipTransferAmount"
    msg = "Tip transfer amount is less than expected"


class InvalidHostTipBalance(ProgramError):
    def __init__(self) -> None:
        super().__init__(6015, "Host tup amount is less than accounted for")

    code = 6015
    name = "InvalidHostTipBalance"
    msg = "Host tup amount is less than accounted for"


class OrderWithinFlashOperation(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6016, "Order within flash operation - all otehr actions are blocked"
        )

    code = 6016
    name = "OrderWithinFlashOperation"
    msg = "Order within flash operation - all otehr actions are blocked"


class CPINotAllowed(ProgramError):
    def __init__(self) -> None:
        super().__init__(6017, "CPI not allowed")

    code = 6017
    name = "CPINotAllowed"
    msg = "CPI not allowed"


class FlashTakeOrderBlocked(ProgramError):
    def __init__(self) -> None:
        super().__init__(6018, "Flash take_order is blocked")

    code = 6018
    name = "FlashTakeOrderBlocked"
    msg = "Flash take_order is blocked"


class FlashTxWithUnexpectedIxs(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6019,
            "Some unexpected instructions are present in the tx. Either before or after the flash ixs, or some ix target the same program between",
        )

    code = 6019
    name = "FlashTxWithUnexpectedIxs"
    msg = "Some unexpected instructions are present in the tx. Either before or after the flash ixs, or some ix target the same program between"


class FlashIxsNotEnded(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6020, "Flash ixs initiated without the closing ix in the transaction"
        )

    code = 6020
    name = "FlashIxsNotEnded"
    msg = "Flash ixs initiated without the closing ix in the transaction"


class FlashIxsNotStarted(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6021, "Flash ixs ended without the starting ix in the transaction"
        )

    code = 6021
    name = "FlashIxsNotStarted"
    msg = "Flash ixs ended without the starting ix in the transaction"


class FlashIxsAccountMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6022, "Some accounts differ between the two flash ixs")

    code = 6022
    name = "FlashIxsAccountMismatch"
    msg = "Some accounts differ between the two flash ixs"


class FlashIxsArgsMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6023, "Some args differ between the two flash ixs")

    code = 6023
    name = "FlashIxsArgsMismatch"
    msg = "Some args differ between the two flash ixs"


class OrderNotWithinFlashOperation(ProgramError):
    def __init__(self) -> None:
        super().__init__(6024, "Order is not within flash operation")

    code = 6024
    name = "OrderNotWithinFlashOperation"
    msg = "Order is not within flash operation"


class EmergencyModeEnabled(ProgramError):
    def __init__(self) -> None:
        super().__init__(6025, "Emergency mode is enabled")

    code = 6025
    name = "EmergencyModeEnabled"
    msg = "Emergency mode is enabled"


class CreatingNewOrdersBlocked(ProgramError):
    def __init__(self) -> None:
        super().__init__(6026, "Creating new ordersis blocked")

    code = 6026
    name = "CreatingNewOrdersBlocked"
    msg = "Creating new ordersis blocked"


class OrderTakingBlocked(ProgramError):
    def __init__(self) -> None:
        super().__init__(6027, "Orders taking is blocked")

    code = 6027
    name = "OrderTakingBlocked"
    msg = "Orders taking is blocked"


class OrderInputAmountTooLarge(ProgramError):
    def __init__(self) -> None:
        super().__init__(6028, "Order input amount larger than the remaining")

    code = 6028
    name = "OrderInputAmountTooLarge"
    msg = "Order input amount larger than the remaining"


class PermissionRequiredPermissionlessNotEnabled(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6029,
            "Permissionless order taking not enabled, please provide permission account",
        )

    code = 6029
    name = "PermissionRequiredPermissionlessNotEnabled"
    msg = "Permissionless order taking not enabled, please provide permission account"


class PermissionDoesNotMatchOrder(ProgramError):
    def __init__(self) -> None:
        super().__init__(6030, "Permission address does not match order address")

    code = 6030
    name = "PermissionDoesNotMatchOrder"
    msg = "Permission address does not match order address"


class InvalidAtaAddress(ProgramError):
    def __init__(self) -> None:
        super().__init__(6031, "Invalid ata address")

    code = 6031
    name = "InvalidAtaAddress"
    msg = "Invalid ata address"


class MakerOutputAtaRequired(ProgramError):
    def __init__(self) -> None:
        super().__init__(6032, "Maker output ata required when output mint is not WSOL")

    code = 6032
    name = "MakerOutputAtaRequired"
    msg = "Maker output ata required when output mint is not WSOL"


class IntermediaryOutputTokenAccountRequired(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6033, "Intermediary output token account required when output mint is WSOL"
        )

    code = 6033
    name = "IntermediaryOutputTokenAccountRequired"
    msg = "Intermediary output token account required when output mint is WSOL"


class NotEnoughBalanceForRent(ProgramError):
    def __init__(self) -> None:
        super().__init__(6034, "Not enough balance for rent")

    code = 6034
    name = "NotEnoughBalanceForRent"
    msg = "Not enough balance for rent"


class NotEnoughTimePassedSinceLastUpdate(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6035, "Order can not be closed - Not enough time passed since last update"
        )

    code = 6035
    name = "NotEnoughTimePassedSinceLastUpdate"
    msg = "Order can not be closed - Not enough time passed since last update"


CustomError = typing.Union[
    OrderCanNotBeCanceled,
    OrderNotActive,
    InvalidAdminAuthority,
    InvalidPdaAuthority,
    InvalidConfigOption,
    InvalidOrderOwner,
    OutOfRangeIntegralConversion,
    InvalidFlag,
    MathOverflow,
    OrderInputAmountInvalid,
    OrderOutputAmountInvalid,
    InvalidHostFee,
    IntegerOverflow,
    InvalidTipBalance,
    InvalidTipTransferAmount,
    InvalidHostTipBalance,
    OrderWithinFlashOperation,
    CPINotAllowed,
    FlashTakeOrderBlocked,
    FlashTxWithUnexpectedIxs,
    FlashIxsNotEnded,
    FlashIxsNotStarted,
    FlashIxsAccountMismatch,
    FlashIxsArgsMismatch,
    OrderNotWithinFlashOperation,
    EmergencyModeEnabled,
    CreatingNewOrdersBlocked,
    OrderTakingBlocked,
    OrderInputAmountTooLarge,
    PermissionRequiredPermissionlessNotEnabled,
    PermissionDoesNotMatchOrder,
    InvalidAtaAddress,
    MakerOutputAtaRequired,
    IntermediaryOutputTokenAccountRequired,
    NotEnoughBalanceForRent,
    NotEnoughTimePassedSinceLastUpdate,
]
CUSTOM_ERROR_MAP: dict[int, CustomError] = {
    6000: OrderCanNotBeCanceled(),
    6001: OrderNotActive(),
    6002: InvalidAdminAuthority(),
    6003: InvalidPdaAuthority(),
    6004: InvalidConfigOption(),
    6005: InvalidOrderOwner(),
    6006: OutOfRangeIntegralConversion(),
    6007: InvalidFlag(),
    6008: MathOverflow(),
    6009: OrderInputAmountInvalid(),
    6010: OrderOutputAmountInvalid(),
    6011: InvalidHostFee(),
    6012: IntegerOverflow(),
    6013: InvalidTipBalance(),
    6014: InvalidTipTransferAmount(),
    6015: InvalidHostTipBalance(),
    6016: OrderWithinFlashOperation(),
    6017: CPINotAllowed(),
    6018: FlashTakeOrderBlocked(),
    6019: FlashTxWithUnexpectedIxs(),
    6020: FlashIxsNotEnded(),
    6021: FlashIxsNotStarted(),
    6022: FlashIxsAccountMismatch(),
    6023: FlashIxsArgsMismatch(),
    6024: OrderNotWithinFlashOperation(),
    6025: EmergencyModeEnabled(),
    6026: CreatingNewOrdersBlocked(),
    6027: OrderTakingBlocked(),
    6028: OrderInputAmountTooLarge(),
    6029: PermissionRequiredPermissionlessNotEnabled(),
    6030: PermissionDoesNotMatchOrder(),
    6031: InvalidAtaAddress(),
    6032: MakerOutputAtaRequired(),
    6033: IntermediaryOutputTokenAccountRequired(),
    6034: NotEnoughBalanceForRent(),
    6035: NotEnoughTimePassedSinceLastUpdate(),
}


def from_code(code: int) -> typing.Optional[CustomError]:
    maybe_err = CUSTOM_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
from ..program_id import PROGRAM_ID


class SetRelayerAccounts(typing.TypedDict):
    admin: Pubkey
    express_relay_metadata: Pubkey
    relayer_signer: Pubkey
    fee_receiver_relayer: Pubkey


def set_relayer(
    accounts: SetRelayerAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
        AccountMeta(
            pubkey=accounts["express_relay_metadata"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["relayer_signer"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["fee_receiver_relayer"], is_signer=False, is_writable=False
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x17\xf3!XnT\xc4%"
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from enum import Enum
from typing import Annotated
from uuid import UUID

from pydantic import PlainSerializer


class UnsupportedOpportunityVersionException(Exception):
    pass


class UnsupportedOpportunityDeleteVersionException(Exception):
    pass


class UnsupportedOpportunityDeleteChainTypeException(Exception):
    pass


class BidStatusVariantsEvm(Enum):
    PENDING = "pending"
    SUBMITTED = "submitted"
    LOST = "lost"
    WON = "won"


class BidStatusVariantsSvm(Enum):
    PENDING = "pending"
    SUBMITTED = "submitted"
    LOST = "lost"
    WON = "won"
    FAILED = "failed"
    EXPIRED = "expired"


IntString = Annotated[int, PlainSerializer(lambda x: str(x), return_type=str)]
UUIDString = Annotated[UUID, PlainSerializer(lambda x: str(x), return_type=str)]
